package com.alicloud.openservices.tablestore.model;

import com.alicloud.openservices.tablestore.ClientException;
import com.alicloud.openservices.tablestore.core.utils.Jsonizable;
import com.alicloud.openservices.tablestore.core.utils.OptionalValue;
import com.alicloud.openservices.tablestore.core.utils.Preconditions;

public class SSESpecification implements Jsonizable {

    /**
     * 是否开启服务器端加密
     */
    private boolean enable = false;

    /**
     * 当开启服务器端加密时，该参数用于设置秘钥类型
     */
    private OptionalValue<SSEKeyType> keyType = new OptionalValue<SSEKeyType>("KeyType");

    /**
     * 构造一个SSESpecification对象。
     * 注意：传入的enable必须为false，原因是当enableStream为true时，必须指定keyType。
     *      如果需要设置开启服务器端加密，请使用另一构造函数。
     * @param enable 必须为false，代表关闭服务器端加密
     */
    public SSESpecification(boolean enable) {
        if (enable) {
            throw new ClientException("Key type is required when enable is true.");
        }
        setEnable(enable);
    }

    /**
     * 构造一个SSESpecification对象。
     * 注意：传入的enable必须为true，原因是当enable为false时，不能指定keyType。
     *      如果需要设置关闭服务器端加密，请使用另一构造函数。
     * @param enable 必须为true，代表开启服务器端加密
     * @param keyType 可选值：SSE_KMS_SERVICE 
     */
    public SSESpecification(boolean enable, SSEKeyType keyType) {
        if (!enable) {
            throw new ClientException("Key type cannot be set when enable is false.");
        }
        setEnable(enable);
        setKeyType(keyType);
    }

    /**
     * 返回是否开启服务器端加密
     *
     * @return
     */
    public boolean isEnable() {
        return enable;
    }

    /**
     * 设置是否开启服务器端加密
     *
     * @param enable
     */
    public void setEnable(boolean enable) {
        this.enable = enable;
    }

    /**
     * 获取keyType参数
     *
     * @return keyType
     */
    public SSEKeyType getKeyType() {
        return keyType.getValue();
    }

    /**
     * 设置keyType参数
     *
     * @param keyType
     */
    public void setKeyType(SSEKeyType keyType) {
        this.keyType.setValue(keyType);
    }

    @Override
    public String jsonize() {
        StringBuilder sb = new StringBuilder();
        jsonize(sb, "\n  ");
        return sb.toString();
    }

    @Override
    public void jsonize(StringBuilder sb, String newline) {
        sb.append('{');
        sb.append(newline);
        sb.append("\"Enable\": ");
        sb.append(enable);
        if (keyType.isValueSet()) {
            sb.append(",");
            sb.append(newline);
            sb.append("\"KeyType\": ");
            sb.append(keyType.getValue().toString());
            sb.append(newline);
        }
        sb.append("}");
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("Enable: ");
        sb.append(enable);
        if (keyType.isValueSet()) {
            sb.append(", KeyType: ");
            sb.append(keyType.getValue().toString());
        }
        return sb.toString();
    }
}
