package com.alicloud.openservices.tablestore.model.search;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.serializer.SerializerFeature;

import com.alicloud.openservices.tablestore.model.OperationNames;
import com.alicloud.openservices.tablestore.model.PrimaryKey;
import com.alicloud.openservices.tablestore.model.PrimaryKeyBuilder;
import com.alicloud.openservices.tablestore.model.Request;

/**
 * SearchIndex的搜索的Request
 */
public class SearchRequest implements Request {

    public static class ColumnsToGet {
        private List<String> columns;
        private boolean returnAll;

        public List<String> getColumns() {
            return columns;
        }

        public void setColumns(List<String> columns) {
            this.columns = columns;
        }

        public boolean isReturnAll() {
            return returnAll;
        }

        public void setReturnAll(boolean returnAll) {
            this.returnAll = returnAll;
        }
    }

    public SearchRequest() {
    }

    public SearchRequest(String tableName, String indexName, SearchQuery searchQuery) {
        this.tableName = tableName;
        this.indexName = indexName;
        this.searchQuery = searchQuery;
    }

    /**
     * TableStore的表名
     */
    private String tableName;
    /**
     * SearchIndex中的index名
     */
    private String indexName;
    /**
     * 查询语句，具体参数详见{@link SearchQuery}
     */
    private SearchQuery searchQuery;
    /**
     * 指定哪些属性列需要返回
     * <p>如果SearchIndex中的属性列太多，而只想要某些属性列，则可以减少网络传输的数据量，提高响应速度</p>
     */
    private ColumnsToGet columnsToGet;
    /**
     * 路由字段
     * 默认为空，大多数场景下不需要使用该值。如果使用了自定义路由，可以指定路由字段。
     * <p>注意：<b>高级特性</b>。如需了解或使用请提工单或联系开发人员</p>
     */
    private List<PrimaryKey> routingValues;

    public String getTableName() {
        return tableName;
    }

    public void setTableName(String tableName) {
        this.tableName = tableName;
    }

    public String getIndexName() {
        return indexName;
    }

    public void setIndexName(String indexName) {
        this.indexName = indexName;
    }

    public SearchQuery getSearchQuery() {
        return searchQuery;
    }

    public void setSearchQuery(SearchQuery searchQuery) {
        this.searchQuery = searchQuery;
    }

    public ColumnsToGet getColumnsToGet() {
        return columnsToGet;
    }

    public void setColumnsToGet(ColumnsToGet columnsToGet) {
        this.columnsToGet = columnsToGet;
    }

    public List<PrimaryKey> getRoutingValues() {
        return routingValues;
    }

    public void setRoutingValues(List<PrimaryKey> routingValues) {
        this.routingValues = routingValues;
    }

    @Override
    public String getOperationName() {
        return OperationNames.OP_SEARCH;
    }

    public String getRequestInfo(boolean prettyFormat) {
        if (prettyFormat) {
            return JSON.toJSONString(this, SerializerFeature.DisableCircularReferenceDetect,
                SerializerFeature.IgnoreErrorGetter, SerializerFeature.PrettyFormat);
        } else {
            return JSON.toJSONString(this, SerializerFeature.DisableCircularReferenceDetect,
                SerializerFeature.IgnoreErrorGetter);
        }
    }

    public void printRequestInfo() {
        System.out.println(getRequestInfo(true));
    }

    private SearchRequest(Builder builder) {
        setTableName(builder.tableName);
        setIndexName(builder.indexName);
        setSearchQuery(builder.searchQuery);
        setColumnsToGet(builder.columnsToGet);
        setRoutingValues(builder.routingValues);
    }

    public static Builder newBuilder() {
        return new Builder();
    }

    public static final class Builder {
        private String tableName;
        private String indexName;
        private SearchQuery searchQuery;
        private ColumnsToGet columnsToGet;
        private List<PrimaryKey> routingValues;

        private Builder() {}

        public Builder tableName(String tableName) {
            this.tableName = tableName;
            return this;
        }

        public Builder indexName(String indexName) {
            this.indexName = indexName;
            return this;
        }

        public Builder searchQuery(SearchQuery searchQuery) {
            this.searchQuery = searchQuery;
            return this;
        }

        /**
         * 指定哪些属性列需要返回
         * <p>如果SearchIndex中的属性列太多，而只想要某些属性列，则可以减少网络传输的数据量，提高响应速度</p>
         */
        public Builder addColumnsToGet(String... columnsToGet) {
            if (this.columnsToGet == null) {
                this.columnsToGet = new ColumnsToGet();
                this.columnsToGet.columns = new ArrayList<String>();
            }
            this.columnsToGet.getColumns().addAll(Arrays.asList(columnsToGet));
            return this;
        }

        /**
         * 指定哪些属性列需要返回
         */
        public Builder addColumnsToGet(List<String> columnsToGetList) {
            if (this.columnsToGet == null) {
                this.columnsToGet = new ColumnsToGet();
                this.columnsToGet.columns = new ArrayList<String>();
            }
            this.columnsToGet.getColumns().addAll(columnsToGetList);
            return this;
        }

        /**
         * 结果集中返回所有的列，不需要具体指定哪些需要返回。
         * <p>如果SearchIndex中的属性列太多，而只想要某些属性列，则可以减少网络传输的数据量，提高响应速度</p>
         */
        public Builder returnAllColumns(boolean returnAll) {
            if (this.columnsToGet == null) {
                this.columnsToGet = new ColumnsToGet();
            }
            this.columnsToGet.setReturnAll(returnAll);
            return this;
        }

        /**
         * 设置路由信息，需要在创建表时候指定路由信息，可通过多次执行该函数设置多次。推荐资讯官方客服后进行使用。
         *
         * @param primaryKeyBuilder 路由主键信息。
         * @return Builder
         */
        public Builder addRoutingValue(PrimaryKeyBuilder primaryKeyBuilder) {
            if (routingValues == null) {
                routingValues = new ArrayList<PrimaryKey>();
            }
            this.routingValues.add(primaryKeyBuilder.build());
            return this;
        }

        /**
         * 设置路由信息，需要在创建表时候指定路由信息。推荐资讯官方客服后进行使用。
         *
         * @param primaryKeys 路由主键信息列表。
         * @return Builder
         */
        public Builder addRoutingValues(List<PrimaryKey> primaryKeys) {
            if (routingValues == null) {
                routingValues = new ArrayList<PrimaryKey>();
            }
            this.routingValues.addAll(primaryKeys);
            return this;
        }

        public SearchRequest build() {
            return new SearchRequest(this);
        }
    }
}
